use std::cell::RefCell;
use std::rc::Rc;

mod msg;

mod crypto;
mod helper;
mod io;
pub mod service;

pub mod prelude {
    pub use crate::helper::{AsAny, IntoWeak};
    pub use crate::io::Packet as FleckPacket;
    pub use crate::msg::Message;
    pub use crate::service::{
        core as fleck_core, event::Event as FleckEvent, order, ChannelSpec, EventRoot, Service,
    };
}

use prelude::*;

pub struct Fleck {
    io: Rc<io::FleckIO>,
    services: RefCell<service::ServiceStack>,
}

impl Fleck {
    pub fn new() -> Rc<Self> {
        let res = Rc::new(Self {
            io: Default::default(),
            services: Default::default(),
        });

        res.register_core_services();

        res
    }

    fn register_core_services(&self) {
        let mut svcs = self.services.borrow_mut();
        // initial incoming/minor/major channels
        svcs.create_io_channels();
        // Node registration
        svcs.give_service::<service::core::Nodes>();
        // Local discovery
        svcs.give_service::<service::core::LocalDiscovery>();
        // Parsing incoming packets
        svcs.give_service::<msg::ParsePacket>();
        // Actually sending packets
        svcs.give_service::<service::core::SendPacket>();
        // Signing packets
        svcs.give_service::<crypto::SignPacket>();
    }

    pub fn run(&self) {
        self.io.poll(self);
    }
}

impl Fleck {
    pub(crate) fn raw_io(&self) -> &io::FleckIO {
        &self.io
    }
    /// Add a service to the Fleck instance. Must not be invoked from a service context.
    pub fn add_service<S: Service + Default + 'static>(&self)
    where
        Rc<S>: Service,
    {
        self.services.borrow_mut().give_service::<S>();
    }
    pub fn services(&self) -> std::cell::Ref<service::ServiceStack> {
        self.services.borrow()
    }
    pub fn with_service<S: 'static, R, F: FnOnce(&S) -> R>(&self, f: F) -> R
    where
        Rc<S>: Service,
    {
        self.services.borrow().with_service(f)
    }
    pub fn queue<CS: service::ChannelSpec>(&self, data: CS::Data) {
        self.services
            .borrow()
            .event_root()
            .channel::<CS>()
            .queue(data);
    }
    pub fn queue_priority<CS: service::ChannelSpec>(&self, data: CS::Data) {
        self.services
            .borrow()
            .event_root()
            .priority_channel::<CS>()
            .queue(data);
    }
}

impl io::IOFeedback for Fleck {
    fn packet(&self, packet: io::Packet) {
        self.services.borrow().process_incoming(self, packet);
    }

    fn minor_tick(&self) {
        self.services.borrow().process_minor_tick(self);
    }

    fn major_tick(&self) {
        self.services.borrow().process_major_tick(self);
    }
}
