use std::cell::RefCell;
use std::{
    collections::{HashMap, VecDeque},
    rc::{Rc, Weak},
};

use super::priority::AbstractServicePriority;

pub struct Event<Data: 'static> {
    pub data: Data,
    pub emptied: bool,
}

struct ConcreteEventSub<Host, Context: 'static + ?Sized, Data: 'static> {
    host: Weak<Host>,
    callback: Box<dyn Fn(&Host, &Context, &mut Event<Data>)>,
}

trait EventSub<Context: 'static + ?Sized, Data: 'static> {
    fn is_healthy(&self) -> bool;
    fn invoke(&self, context: &Context, data: &mut Event<Data>);
}

impl<Host: 'static, Context: 'static + ?Sized, Data: 'static> EventSub<Context, Data>
    for ConcreteEventSub<Host, Context, Data>
{
    fn is_healthy(&self) -> bool {
        self.host.strong_count() > 0
    }
    fn invoke(&self, context: &Context, data: &mut Event<Data>) {
        self.host
            .upgrade()
            .map(|h| (self.callback)(h.as_ref(), context, data));
    }
}

pub struct EventChannel<Tag: 'static, Context: 'static + ?Sized, Data: 'static> {
    subs: RefCell<Vec<Box<dyn EventSub<Context, Data>>>>,

    queue: RefCell<VecDeque<Data>>,

    _ghost: std::marker::PhantomData<(Tag,)>,
}

impl<Tag: 'static, Context: 'static + ?Sized, Data: 'static> Default
    for EventChannel<Tag, Context, Data>
{
    fn default() -> Self {
        Self {
            subs: RefCell::new(Vec::new()),
            queue: Default::default(),
            _ghost: std::marker::PhantomData,
        }
    }
}

impl<Tag: 'static, Context: 'static + ?Sized, Data: 'static> EventChannel<Tag, Context, Data> {
    pub fn subscribe<
        Host: 'static,
        HC: crate::helper::IntoWeak<Host>,
        CB: Fn(&Host, &Context, &mut Event<Data>) + 'static,
    >(
        &self,
        who: HC,
        cb: CB,
    ) {
        self.subs.borrow_mut().push(Box::new(ConcreteEventSub {
            host: who.as_weak(),
            callback: Box::new(cb),
        }));
    }

    pub fn queue(&self, data: Data) {
        self.queue.borrow_mut().push_back(data);
    }

    fn do_fire_all(&self, context: &Context) -> usize {
        let mut count = 0;

        while !self.queue.borrow().is_empty() {
            let mut event = Event {
                data: self.queue.borrow_mut().pop_front().unwrap(),
                emptied: false,
            };

            for sub in self.subs.borrow().iter() {
                sub.invoke(context, &mut event);
                if event.emptied {
                    break;
                }
            }

            count += 1;
        }

        count
    }
}

pub struct PriorityChannel<Tag: 'static, Context: 'static + ?Sized, Data: 'static> {
    subs: RefCell<super::priority::TotalOrder<Rc<dyn EventSub<Context, Data>>>>,
    order_cache: RefCell<Option<Vec<Rc<dyn EventSub<Context, Data>>>>>,

    queue: RefCell<VecDeque<Data>>,

    _ghost: std::marker::PhantomData<(Tag,)>,
}

impl<Tag: 'static, Context: 'static + ?Sized, Data: 'static> Default
    for PriorityChannel<Tag, Context, Data>
{
    fn default() -> Self {
        Self {
            subs: RefCell::new(Default::default()),
            order_cache: Default::default(),
            queue: RefCell::new(Default::default()),
            _ghost: std::marker::PhantomData,
        }
    }
}

pub struct PrioritySubscriptionBuilder<
    'l,
    Tag: 'static,
    Context: 'static + ?Sized,
    Data: 'static,
    Priority: AbstractServicePriority,
> {
    parent: &'l PriorityChannel<Tag, Context, Data>,
    _ghost: std::marker::PhantomData<Priority>,
}

impl<
        'l,
        Tag: 'static,
        Context: 'static + ?Sized,
        Data: 'static,
        Priority: AbstractServicePriority,
    > PrioritySubscriptionBuilder<'l, Tag, Context, Data, Priority>
{
    pub fn subscribe<
        Host: 'static,
        HC: crate::helper::IntoWeak<Host>,
        CB: Fn(&Host, &Context, &mut Event<Data>) + 'static,
    >(
        &self,
        who: HC,
        cb: CB,
    ) {
        self.parent.subscribe::<Priority, Host, HC, CB>(who, cb);
    }
}

impl<Tag: 'static, Context: 'static + ?Sized, Data: 'static> PriorityChannel<Tag, Context, Data> {
    pub fn subscribe<
        P: super::AbstractServicePriority,
        Host: 'static,
        HC: crate::helper::IntoWeak<Host>,
        CB: Fn(&Host, &Context, &mut Event<Data>) + 'static,
    >(
        &self,
        who: HC,
        cb: CB,
    ) {
        let sub = Rc::new(ConcreteEventSub {
            host: who.as_weak(),
            callback: Box::new(cb),
        });

        self.subs.borrow_mut().add_priority::<P>(sub);
        self.order_cache.borrow_mut().take();
    }

    pub fn with<P: AbstractServicePriority>(
        &self,
    ) -> PrioritySubscriptionBuilder<Tag, Context, Data, P> {
        PrioritySubscriptionBuilder {
            parent: self,
            _ghost: std::marker::PhantomData,
        }
    }

    pub fn queue(&self, data: Data) {
        self.queue.borrow_mut().push_back(data);
    }

    fn do_fire_all(&self, context: &Context) -> usize {
        let mut cache = self.order_cache.borrow_mut();
        let subs = cache.get_or_insert_with(|| self.subs.borrow().clone().order());

        let mut count = 0;

        while !self.queue.borrow().is_empty() {
            let mut event = Event {
                data: self.queue.borrow_mut().pop_front().unwrap(),
                emptied: false,
            };

            for sub in subs.iter() {
                sub.invoke(context, &mut event);
                if event.emptied {
                    break;
                }
            }

            count += 1;
        }

        count
    }
}

trait ChannelMetadata<Context: 'static + ?Sized> {
    fn queue_len(&self) -> usize;
    fn fire_all(&self, context: &Context) -> usize;
}

impl<Tag: 'static, Context: 'static + ?Sized, Data: 'static> ChannelMetadata<Context>
    for EventChannel<Tag, Context, Data>
{
    fn queue_len(&self) -> usize {
        self.queue.borrow().len()
    }

    fn fire_all(&self, context: &Context) -> usize {
        self.do_fire_all(context)
    }
}

impl<Tag: 'static, Context: 'static + ?Sized, Data: 'static> ChannelMetadata<Context>
    for PriorityChannel<Tag, Context, Data>
{
    fn queue_len(&self) -> usize {
        self.queue.borrow().len()
    }

    fn fire_all(&self, context: &Context) -> usize {
        self.do_fire_all(context)
    }
}

pub trait ChannelSpec {
    type Tag: 'static;
    type Data: 'static;
}

impl<Tag: 'static, Data: 'static> ChannelSpec for (Tag, Data) {
    type Tag = Tag;
    type Data = Data;
}

pub struct EventRoot<Context: 'static + ?Sized> {
    channels: HashMap<std::any::TypeId, Rc<dyn std::any::Any>>,
    metadata: HashMap<std::any::TypeId, (&'static str, Rc<dyn ChannelMetadata<Context>>)>,
    _ghost: std::marker::PhantomData<Context>,
}

impl<Context: 'static + ?Sized> Default for EventRoot<Context> {
    fn default() -> Self {
        Self {
            channels: Default::default(),
            metadata: Default::default(),
            _ghost: std::marker::PhantomData,
        }
    }
}

impl<Context: 'static + ?Sized> EventRoot<Context> {
    pub fn create_channel<CS: ChannelSpec>(&mut self) {
        let tid = std::any::TypeId::of::<EventChannel<CS::Tag, Context, CS::Data>>();

        let ch = Rc::new(EventChannel::<CS::Tag, Context, CS::Data>::default());
        self.metadata
            .insert(tid, (std::any::type_name::<CS>(), ch.clone()));
        self.channels.insert(tid, ch);
    }

    pub fn create_priority_channel<CS: ChannelSpec>(&mut self) {
        let tid = std::any::TypeId::of::<PriorityChannel<CS::Tag, Context, CS::Data>>();

        let ch = Rc::new(PriorityChannel::<CS::Tag, Context, CS::Data>::default());
        self.metadata
            .insert(tid, (std::any::type_name::<CS>(), ch.clone()));
        self.channels.insert(tid, ch);
    }

    pub fn channel<CS: ChannelSpec>(&self) -> &EventChannel<CS::Tag, Context, CS::Data> {
        let tid = std::any::TypeId::of::<EventChannel<CS::Tag, Context, CS::Data>>();

        match self.channels.get(&tid) {
            Some(ch) => ch.downcast_ref().expect("internal inconsistency"),
            None => {
                panic!(
                    "Asked for channel {} that has not been created!",
                    std::any::type_name::<CS>()
                )
            }
        }
    }

    pub fn priority_channel<CS: ChannelSpec>(
        &self,
    ) -> &PriorityChannel<CS::Tag, Context, CS::Data> {
        let tid = std::any::TypeId::of::<PriorityChannel<CS::Tag, Context, CS::Data>>();

        match self.channels.get(&tid) {
            Some(ch) => ch.downcast_ref().expect("internal inconsistency"),
            None => {
                if self.channels.contains_key(&std::any::TypeId::of::<
                    EventChannel<CS::Tag, Context, CS::Data>,
                >()) {
                    panic!("Asked for priority channel {}, which does not exist, but there is a non-priority channel!", std::any::type_name::<CS>())
                } else {
                    panic!(
                        "Asked for priority channel {} that has not been created!",
                        std::any::type_name::<CS>()
                    )
                }
            }
        }
    }

    pub(crate) fn fire_all(&self, context: &Context) {
        let mut any = true;
        while any {
            any = false;

            for ch in &self.metadata {
                let count = ch.1 .1.fire_all(context);
                if count > 0 {
                    log::trace!("Queue {} processed {} event(s)", ch.1 .0, count);
                }
                any = any || count > 0;
            }
        }
    }
}

#[cfg(test)]
mod tests {
    use std::ops::Deref;
    use std::{
        cell::{Cell, RefCell},
        rc::Rc,
    };

    use super::{Event, EventRoot};
    use crate::service::order;

    struct Receiver {
        int_count: Cell<usize>,
    }

    impl Receiver {
        fn receive_i32(&self, _ctx: &(), _val: &mut Event<i32>) {
            self.int_count.set(self.int_count.get() + 1);
        }
    }

    struct OrderedReceiver {
        id: i32,
        order: Rc<RefCell<Vec<i32>>>,
    }

    impl OrderedReceiver {
        fn receive_i32(&self, _ctx: &(), _val: &mut Event<i32>) {
            self.order.borrow_mut().push(self.id)
        }
    }

    struct IntTag;
    type IntChannel = (IntTag, i32);

    #[test]
    fn simple_fire() {
        let mut root = EventRoot::default();

        root.create_channel::<IntChannel>();

        let recv = Rc::new(Receiver {
            int_count: Cell::new(0),
        });

        root.channel::<IntChannel>()
            .subscribe(&recv, Receiver::receive_i32);
        root.channel::<IntChannel>().queue(0i32);
        assert_eq!(recv.int_count.get(), 0);
        root.channel::<IntChannel>().do_fire_all(&());
        assert_eq!(recv.int_count.get(), 1);
    }

    #[test]
    fn priority_fire() {
        let mut root = EventRoot::default();

        root.create_priority_channel::<IntChannel>();

        let order = Rc::new(RefCell::new(Vec::new()));

        let recv1 = Rc::new(OrderedReceiver {
            id: 1,
            order: order.clone(),
        });
        let recv2 = Rc::new(OrderedReceiver {
            id: 2,
            order: order.clone(),
        });

        root.priority_channel::<IntChannel>()
            .with::<order::First>()
            .subscribe(&recv1, OrderedReceiver::receive_i32);
        root.priority_channel::<IntChannel>()
            .with::<order::Last>()
            .subscribe(&recv2, OrderedReceiver::receive_i32);
        root.priority_channel::<IntChannel>().queue(0i32);
        assert_eq!(order.borrow().deref(), &vec![]);
        root.priority_channel::<IntChannel>().do_fire_all(&());
        assert_eq!(order.borrow().deref(), &vec![1, 2]);
    }
}
