use serde::{Deserialize, Serialize};
use std::rc::Rc;

use super::core::MajorTickChannel;
use super::event::Event;
use super::{order, BetweenPriority, Service};
use crate::io::Packet;

use crate as fleck;
use crate::msg::MessageParams;

pub struct SendPacketTag {}
pub type SendPacketChannel = (SendPacketTag, crate::io::Packet);

#[derive(Default)]
pub(crate) struct SendPacket {}

impl Service for Rc<SendPacket> {
    fn register_channels(&self, eroot: &mut super::EventRoot) {
        eroot.create_priority_channel::<SendPacketChannel>();
        eroot
            .priority_channel::<SendPacketChannel>()
            .with::<order::Last>()
            .subscribe(self, SendPacket::process_outgoing);
    }
}

impl SendPacket {
    fn process_outgoing(&self, api: &crate::Fleck, packet: &mut Event<crate::io::Packet>) {
        // use default channel if not specified
        if packet.data.io_channel.is_none() {
            packet.data.io_channel = Some(api.raw_io().global().clone());
        }

        // serialize if needed
        if let (false, Some(msg)) = (packet.data.data.as_ref().is_some(), packet.data.msg.take()) {
            packet.data.data = Some(bincode::serialize(&msg).expect("failed to serialize message"));
        }

        match &mut packet.data.io_channel {
            Some(ch) => ch.clone().send_packet(&mut packet.data),
            _ => {
                println!("tried to send packet with no channel?");
            }
        }
    }
}

#[derive(Default)]
pub(crate) struct LocalDiscovery;

#[derive(Serialize, Deserialize, Debug)]
struct DiscoveryMsg {
    pkey: crate::crypto::PublicKey,
}

impl MessageParams for DiscoveryMsg {
    const NAME: &'static str = "DiscoveryMsg";
    const ENCRYPTED: bool = false;
}

impl Service for Rc<LocalDiscovery> {
    fn register_msgs(&self, registry: &mut fleck::msg::MessageRegistry) {
        registry.add_message_type::<DiscoveryMsg>();
    }

    fn register_channels(self: &Self, eroot: &mut super::EventRoot) {
        eroot
            .channel::<MajorTickChannel>()
            .subscribe(self, LocalDiscovery::process_major_tick);
        eroot
            .priority_channel::<super::core::ReceivePacketChannel>()
            .with::<order::Processing>()
            .subscribe(self, LocalDiscovery::packet);
    }
}

impl LocalDiscovery {
    fn packet(&self, api: &crate::Fleck, pkt: &mut Event<Packet>) {
        if let Some(discovery) = pkt.data.as_msg::<DiscoveryMsg>() {
            log::trace!("received discovery message");

            // this will automatically ignore self-messages because we already know the pubkey
            api.with_service(|nodes: &super::core::Nodes| {
                if nodes.node_by_pubkey(&discovery.pkey).is_none() {
                    let node = super::core::Node::build_with_addr(pkt.data.addr.unwrap());
                    node.set_pubkey(discovery.pkey);
                    nodes.inform_of(node);
                }
            });
        }
    }

    fn process_major_tick(&self, api: &crate::Fleck, _: &mut Event<()>) {
        api.queue_priority::<SendPacketChannel>(Packet {
            addr: Some((*crate::io::MULTICAST_ADDRESS).into()),
            data: None,
            io_channel: Some(api.raw_io().local().clone()),
            node: None,
            msg: Some(crate::msg::Message::build(DiscoveryMsg {
                pkey: api.with_service(|nodes: &super::core::Nodes| nodes.self_node().pubkey().unwrap().to_owned())
            })),
        });
    }
}
